/***********************************************************************************************************************
 *
 * BlueBook - use live code examples in your book
 * Copyright (C) 2011-2011 by Tidalwave s.a.s.
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebook.java.net
 * SCM: https://hg.java.net/hg/bluebook~src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebook.filter.java;

import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.List;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.Writer;
import java.util.regex.Pattern;
import lombok.Getter;
import lombok.Setter;

/***********************************************************************************************************************
 *
 * TODO: should be replaced by a better implementation based on a Java lexical parser
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 * @stable
 *
 **********************************************************************************************************************/
public class JavaFilter 
  {
    @Getter @Setter
    private boolean skipHeader;
    
    // FIXME: this actually works only in some simplified cases
    @Getter @Setter
    private boolean skipBlockComments;
    
    @Getter @Setter
    private List<String> includedFragments = new ArrayList<String>();
    
    @Getter @Setter
    private List<String> excludedFragments = new ArrayList<String>();
    
    @Getter @Setter
    private boolean skipPackage;
    
    @Getter @Setter
    private boolean skipImports;
    
    @Getter @Setter
    private String skipRegexp = "";
    
    public void filter (final @Nonnull Reader r, final @Nonnull Writer w)
      throws IOException
      {
        final BufferedReader br = new BufferedReader(r);
        final PrintWriter pw = new PrintWriter(w);
        boolean inHeader = true;
        boolean inBlockComment = true;
        boolean inIncludedFragment = false;
        boolean inExcludedFragment = false;
        boolean inIgnoredFragment = false;
        boolean prevIsBlank = true;
        
        for (;;)
          {
            final String line = br.readLine();
            
            if (line == null)
              {
                break;  
              }
            
            final String trimmedLine = line.trim();
            
            inHeader &= !line.startsWith("package ");
            inBlockComment |= trimmedLine.startsWith("/*");
            inIncludedFragment |= matches(line, includedFragments, "//@bluebook-begin");
            inExcludedFragment |= matches(line, excludedFragments, "//@bluebook-begin");            
            inIgnoredFragment |= line.contains("//@bluebook-ignore-begin");
            
            boolean skip = false;
            skip |= skipHeader && inHeader;
            skip |= skipBlockComments && inBlockComment;
            skip |= line.contains("//@bluebook");
            skip |= !includedFragments.isEmpty() && !inIncludedFragment;
            skip |= !excludedFragments.isEmpty() && inExcludedFragment;
            skip |= inIgnoredFragment;
            skip |= skipPackage && line.startsWith("package");
            skip |= skipImports && line.startsWith("import ");
            skip |= !"".equals(skipRegexp) && Pattern.matches(skipRegexp, line);
            
            final boolean isBlank = trimmedLine.equals("");
            skip |= isBlank && prevIsBlank;
            
            if (!skip)
              {
                pw.println(line);
                prevIsBlank = isBlank;
              }
            
            inBlockComment &= !trimmedLine.endsWith("*/");            
            inIncludedFragment &= !matches(line, includedFragments, "//@bluebook-end");            
            inExcludedFragment &= !matches(line, excludedFragments, "//@bluebook-end");
            inIgnoredFragment &= !line.contains("//@bluebook-ignore-end");
          }
      }

    private boolean matches (final @Nonnull String line, 
                             final @Nonnull List<String> fragments,
                             final @Nonnull String marker) 
      {
        final int i = line.indexOf(marker + " ");
        
        if (i < 0)
          {
            return false;  
          }
        
        final String s = line.substring(i + marker.length()).trim();
        
        for (final String f : s.split(","))
          {
            if (fragments.contains(f.trim()))
              {
                return true;  
              }
          }
        
        return false;
      }
  }
